#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <sys/wait.h>
#include "dns.h"
#include "socket.h"
#include "strerr.h"
#include "buffer.h"
#include "ip6.h"
#include "timeoutconn.h"
#include "uint16.h"
#include "str.h"
#include "iopause.h"
#include "error.h"
#include "fmt.h"
#include "sgetopt.h"
#include "env.h"
#include "pathexec.h"
#include "byte.h"
#include "ndelay.h"
#include "fd.h"

#define FATAL "ncp: fatal: "

static char seed[128];

char mc6ip[16]={0xff,2,0,0,0,0,0,0,0,0,0,0,'n','c','p','0'};
char mc4ip[4]={224,'n','c','p'};
char bcip[4]={0xff,0xff,0xff,0xff};

static char message[]="libdjb-ncp-1.2.2";

int nano_timeoutaccept6(int s,char ip[16],uint16 *port,unsigned int timeout)
{
  struct taia now;
  struct taia deadline;
  iopause_fd x;
  int fd;

  if ((fd=socket_accept6(s,ip,port,0)) == -1) {
    if ((errno != error_wouldblock) && (errno != error_inprogress)) return -1;
    x.fd = s;
    x.events = IOPAUSE_WRITE;
    taia_now(&now);
    deadline.sec.x=0;
    deadline.nano=timeout;
    deadline.atto=0;
    taia_add(&deadline,&now,&deadline);
    for (;;) {
      taia_now(&now);
      iopause(&x,1,&deadline,&now);
      if (x.revents) break;
      if (taia_less(&deadline,&now)) {
	errno = error_timeout; /* note that connect attempt is continuing */
	return -1;
      }
    }
    if ((fd=socket_accept6(s,ip,port,0)) == -1) return -1;
  }
  return fd;
}

#if 0
long int socket_sendfile(int destfd,int sourcefd,long int maxcopy) {
  char buf[16*1024];
  int len,written;
  long int done=0;
  while (maxcopy-done>0) {
    if (maxcopy-done>16*1024)
      len=read(sourcefd,buf,16*1024);
    else
      len=read(sourcefd,buf,maxcopy-done);
    len=read(sourcefd,buf,16*1024);
    if (len<0) return -1;
    if (len==0) return done;
    written=write(destfd,buf,len);
    if (written<0) return -1;
    if (written<len) return done+written;
    done+=written;
  }
  return done;
}
#endif

void npush(int argc,char *argv[]) {
  int fd;
  int s;
  char ip[16];
  char **newargv;
  int i;
  char *ips[4];
  int fds[4];
  int v6only=0;
  int v4only=0;
  int forcebc=0;
  int opt;

  if (*argv) {
    if (str_equal(*argv,"--usage") || str_equal(*argv,"--help"))
      strerr_die1x(0,"usage: npush [filename...]\nWill send stdin when no filenames are given");
    while ((opt=getopt(argc,argv-1,"64b")) != opteof) {
      switch(opt) {
      case '6': v6only=1; break;
      case '4': v4only=1; break;
      case 'b': forcebc=1; break;
      }
    }
    argv+=optind-1;
  }
  fds[0]=fds[1]=fds[2]=-1;
  if (!v4only && !forcebc) {
    if ((fds[0]=socket_udp6()) == -1)
      strerr_die2sys(111,FATAL,"unable to IPv6 multicast create socket: ");
    if (socket_bind6_reuse(fds[0],(char *)V6any,8002,0) == -1)
      strerr_die2sys(111,FATAL,"unable to bind: ");
    if (!noipv6) {	/* try ipv6 multicast */
      ips[0]=mc6ip;
      if (socket_mcloop6(fds[0],1) == -1) {
	strerr_warn1("npush: IPv6 multicast failed, trying IPv4 multicast",0);
	close(fds[0]); fds[0]=-1;
      }
    } else
      fds[0]=-1;
  }

  if (!v6only && !forcebc) {
    if ((fds[1]=socket_udp()) == -1)
      strerr_die2sys(111,FATAL,"unable to create IPv4 multicast socket: ");
    if (socket_bind4_reuse(fds[1],(char *)V6any,8002) == -1)
      strerr_die2sys(111,FATAL,"unable to bind: ");
    ips[1]=mc4ip;
    if (socket_mcloop4(fds[1],1) == -1) {
      strerr_warn1("npush: IPv4 multicast failed, trying IPv4 broadcast",0);
      close(fds[1]); fds[1]=-1;
    }
  }

  if (forcebc || (fds[0]==0 && fds[1]==0)) {
    if ((fds[2]=socket_udp()) == -1)
      strerr_die2sys(111,FATAL,"unable to create IPv4 broadcast socket: ");
    if (socket_bind4_reuse(fds[2],(char *)V6any,8002) == -1)
      strerr_die2sys(111,FATAL,"unable to bind: ");
    if (socket_broadcast(fds[2]) == -1)
      strerr_die2sys(111,FATAL,"unable to get broadcasting permission: ");
    ips[2]=bcip;
  } else
    fds[2]=-1;

  if ((s=socket_tcp6())==-1)
    strerr_die2sys(111,FATAL,"unable to create TCP socket: ");
  if (socket_bind6_reuse(s,(char *)V6any,8002,0) == -1)
    strerr_die2sys(111,FATAL,"unable to bind: ");
  if (socket_listen(s,1) == -1)
    strerr_die2sys(111,FATAL,"unable to listen: ");

  for (;;) {
    uint16 port;
    if (fds[0]!=-1)
      if (socket_send6(fds[0],message,str_len(message),mc6ip,8002,0)==-1)
	strerr_die2sys(111,FATAL,"IPv6 multicast sendmsg failed: ");
    if (fds[1]!=-1)
      if (socket_send4(fds[1],message,str_len(message),mc4ip,8002)==-1)
	strerr_die2sys(111,FATAL,"IPv4 multicast sendmsg failed: ");
    if (fds[2]!=-1)
      if (socket_send4(fds[2],message,str_len(message),bcip,8002)==-1)
	strerr_die2sys(111,FATAL,"IPv4 broadcast sendmsg failed: ");
    if ((fd=nano_timeoutaccept6(s,ip,&port,200000))!=-1)
      break;
  }
  if (fds[0]!=-1) close(fds[0]);
  if (fds[1]!=-1) close(fds[1]);
  if (fds[2]!=-1) close(fds[2]);
  ndelay_off(fd);
  close(s);
  if (!*argv) {	/* if command line is empty, send \0 and stdin */
    char c=0;
    char buf[1500];
    int len,written;
    if (write(fd,&c,1)!=1)
      strerr_die2sys(111,FATAL,"write failed: ");
    do {
      len=read(0,buf,1500);
      if (len==-1)
	strerr_die2sys(111,FATAL,"read failed: ");
      if (len==0)
	break;
      written=write(fd,buf,len);
      if (written==-1)
	strerr_die2sys(111,FATAL,"write failed: ");
      if (written<len)
	strerr_die2x(111,FATAL,"short write");
    } while (1);
    close(fd);
    exit(0);
  }
  close(1);
  close(0);
  if (fd_move(1,fd) == -1)
    strerr_die2sys(111,FATAL,"unable to set up stdout: ");
  newargv=(char **)malloc((argc+2)*sizeof(char*));
  if (!newargv)
    strerr_die2x(111,FATAL,"out of memory");
  newargv[0]="tar";
  newargv[1]="cpvvf";
  newargv[2]="-";
  i=3;
  while (*argv)
    newargv[i++]=*argv++;
  newargv[i]=0;
  pathexec(newargv);
  strerr_die2sys(111,FATAL,"unable to run tar: ");
  exit(0);
}

void npoll(int argc,char *argv[]) {
  /* TODO: look for UDP packets, connect to sending IP.
    * read 1 byte.  If it is \0, dump the rest to stdout.
    * Otherwise, create a pipe to tar x, write the byte followed by the
    * socket data to the pipe. */
  char buf[1600];
  char ip[16];
  char portstr[10];
  uint16 port;
  char **newargv;
  int pipefd[2];
  int s;
  pid_t pid;
  stralloc out={0};
  int outidx=0;
  uint32 scope_id=0;

  if (*argv && (str_equal(*argv,"--usage") || str_equal(*argv,"--help")))
    strerr_die1x(0,"usage: npoll [hostname...]\nWith hostname, will retrieve files/stdin from hostname.\n"
		 "Without hostname, will look for announcement packets from npush to find sender.");
  if (*argv) {
    stralloc fqdn={0};
    stralloc tmp={0};

    if (!stralloc_copys(&tmp,*argv))
      strerr_die2x(111,FATAL,"out of memory");
    if (dns_ip6_qualify(&out,&fqdn,&tmp) == -1)
      strerr_die4sys(111,FATAL,"temporary unable to figure out IP address for ",*argv,": ");
    if (out.len < 16)
      strerr_die3x(111,FATAL,"no IP address for ",*argv);
    byte_copy(ip,16,out.s);
    buf[ip6_fmt(buf,ip)]=0;
    strerr_warn2("connecting to ",buf,0);
  } else {
    if ((s=socket_udp6())==-1)
      strerr_die2sys(111,FATAL,"could not create socket: ");
    if (socket_bind6_reuse(s,(char *)V6any,8002,0)==-1)
      strerr_die2sys(111,FATAL,"could not bind: ");
    socket_mcjoin6(s,mc6ip,0);
    socket_mcjoin4(s,mc4ip,(char *)V6any);
    ndelay_off(s);
    for (;;) {
      if (socket_recv6(s,buf,1550,ip,&port,&scope_id)==-1)
	strerr_die2sys(111,FATAL,"could not receive UDP packet: ");
      if (port == 8002)
	break;
    }
    buf[ip6_fmt(buf,ip)]=0;
    portstr[fmt_ulong(portstr,port)]=0;
    if (scope_id)
      strerr_warn6("got packet from ",buf," port ",portstr," on interface ",socket_getifname(scope_id),0);
    else
      strerr_warn4("got packet from ",buf," port ",portstr,0);
    close(s);
  }

  for (;;) {
    s = socket_tcp6();
    if (s==-1)
      strerr_die2sys(111,FATAL,"unable to create socket: ");
    if (socket_bind6(s,(char *)V6any,0,0) == -1)
      strerr_die2sys(111,FATAL,"unable to bind socket: ");
    ndelay_off(s);
    if (socket_connect6(s,ip,8002,scope_id) == -1) {
/*    if (timeoutconn6(s,ip,8002,scope_id,60) == -1) { */
      if (out.len>outidx+16) {
	outidx+=16;
	byte_copy(ip,16,out.s+outidx);
	buf[ip6_fmt(buf,ip)]=0;
	strerr_warn2("connecting to ",buf,0);
	close(s);
	continue;
      }
      strerr_die4sys(111,FATAL,"unable to connect to ",*argv,": ");
    } else
      break;
  }
  ndelay_off(s);
  if (read(s,buf,1)!=1)
    strerr_die2sys(111,FATAL,"unable to read: ");
  if (buf[0]==0) {	/* stdin/stdout mode */
    int len,written;
    do {
      len=read(s,buf,1500);
      if (len==-1)
	strerr_die2sys(111,FATAL,"read failed: ");
      if (len==0)
	break;
      written=write(1,buf,len);
      if (written==-1)
	strerr_die2sys(111,FATAL,"write failed: ");
      if (written<len)
	strerr_die2x(111,FATAL,"short write");
    } while (1);
    close(s);
    exit(0);
  }
  /* unfortunately, since we read the first byte, we can't just exec
   * tar now, we need to create a pipe, fork, exec tar and copy the
   * bytes. */
  if (pipe(pipefd)==-1)
    strerr_die2sys(111,FATAL,"pipe failed: ");
  switch (pid=fork()) {
  case 0:
    close(pipefd[1]);
    close(0);
    if (fd_move(0,pipefd[0])==-1)
      strerr_die2sys(111,FATAL,"unable to set up stdin: ");
    newargv=(char **)malloc(4*sizeof(char *));
    if (!newargv)
      strerr_die2x(111,FATAL,"out of memory");
#if 1
    newargv[0]="tar";
    newargv[1]="xvvpf";
    newargv[2]="-";
    newargv[3]=0;
#else
    newargv[0]="strace";
    newargv[1]="tar";
    newargv[2]="xvvpf";
    newargv[3]="-";
    newargv[4]=0;
#endif
    pathexec(newargv);
    strerr_die2sys(111,FATAL,"unable to run tar: ");

  case -1:
    strerr_die2sys(111,FATAL,"can't fork: ");
  }
  close(pipefd[0]);
  ndelay_off(1);
  if (write(pipefd[1],buf,1)==-1)
    strerr_die2sys(111,FATAL,"write failed: ");
  {
    int len,written;
    do {
      len=read(s,buf,1500);
      if (len==-1)
	strerr_die2sys(111,FATAL,"read failed: ");
      if (len==0)
	break;
      written=write(pipefd[1],buf,len);
      if (written==-1)
	strerr_die2sys(111,FATAL,"write failed: ");
      if (written<len)
	strerr_die2x(111,FATAL,"short write");
    } while (1);
  }
  close(s);
  close(pipefd[1]);
  waitpid(pid,0,0);
  exit(0);
}

void ncpserver(int argc,char *argv[]) {
  char remoteip[16];
  int fd,s;
  uint16 remoteport;
  char **newargv;

  strerr_warn1("ncp: server mode.  waiting for connection.",0);
  s = socket_tcp6();
  if (s == -1)
    strerr_die2sys(111,FATAL,"unable to create socket: ");
  if (socket_bind6_reuse(s,(char *)V6any,8002,0) == -1)
    strerr_die2sys(111,FATAL,"unable to bind: ");
  if (socket_listen(s,1) == -1)
    strerr_die2sys(111,FATAL,"unable to listen: ");
  ndelay_off(s);
  if ((fd=socket_accept6(s,remoteip,&remoteport,0)) == -1)
    strerr_die2sys(111,FATAL,"unable to accept: ");
  close(0);
  if (fd_move(0,fd) == -1)
    strerr_die2sys(111,FATAL,"unable to set up stdin: ");
  newargv=(char **)malloc(3*sizeof(char *));
  if (!newargv)
    strerr_die2x(111,FATAL,"out of memory");
  newargv[0]="tar";
  newargv[1]="xvvpf";
  newargv[2]="-";
  newargv[3]=0;
  pathexec(newargv);
  strerr_die2sys(111,FATAL,"unable to run tar: ");
}

void ncpclient(int argc,char *argv[]) {
  int s,i;
  char **newargv;
  stralloc fqdn={0};
  stralloc out={0};
  stralloc tmp={0};

  if (str_equal(*argv,"--usage") || str_equal(*argv,"--help"))
    strerr_die1x(0,"usage: server mode: ncp hostname filename [filename...]\n"
		   "       client mode: ncp");
  if (!stralloc_copys(&tmp,*argv))
    strerr_die2x(111,FATAL,"out of memory");
  if (dns_ip6_qualify(&out,&fqdn,&tmp) == -1)
    strerr_die4sys(111,FATAL,"temporary unable to figure out IP address for ",*argv,": ");
  if (out.len < 16)
    strerr_die3x(111,FATAL,"no IP address for ",*argv);
  s = socket_tcp6();
  if (s==-1)
    strerr_die2sys(111,FATAL,"unable to create socket: ");
  if (socket_bind6(s,(char *)V6any,0,0) == -1)
    strerr_die2sys(111,FATAL,"unable to bind socket: ");
  if (timeoutconn6(s,out.s,8002,0,60) == -1)
    strerr_die4sys(111,FATAL,"unable to connect to ",*argv,": ");
  ndelay_off(s);
  close(1);
  if (fd_move(1,s) == -1)
    strerr_die2sys(111,FATAL,"unable to set up stdout: ");
  newargv=(char **)malloc((argc+2)*sizeof(char *));
  if (!newargv)
    strerr_die2x(111,FATAL,"out of memory");
  newargv[0]="tar";
  newargv[1]="cpvvf";
  newargv[2]="-";
  i=3;
  argv++;
  while (*argv)
    newargv[i++]=*argv++;
  newargv[i]=0;
  pathexec(newargv);
  strerr_die2sys(111,FATAL,"unable to run tar: ");
}

int main(int argc,char *argv[]) {
  char *argv0=*argv;

  dns_random_init(seed);
  if (!argv0)
    strerr_die2sys(111,FATAL,"no argv[0]?!");
  argv0+=str_rchr(argv0,'/');
  if (*argv0=='/') argv0++; else argv0=*argv;
  argv++;
  if (*argv && str_equal(*argv,"--version"))
    strerr_die2x(0,message," (libdjb)\nWritten by Felix von Leitner <felix-ncp@fefe.de>\n"
		     "This is free software; copy and use it as you like. No warranty.\n");
  if (str_equal(argv0,"npush"))
    npush(argc,argv);
  else if (str_equal(argv0,"npoll"))
    npoll(argc,argv);
  else if (!*argv)
    ncpserver(argc,argv);
  else
    ncpclient(argc,argv);
  return 0;
}

